package org.processmining.starstar.model;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.processmining.starstar.constants.ModelConstants;
import org.processmining.starstar.elements.XOCActivity;
import org.processmining.starstar.elements.XOCObject;
import org.processmining.starstar.elements.XOCObjectClass;

public class ModelEdge {
	XOCActivity sourceActivity;
	XOCActivity targetActivity;
	XOCObjectClass relationClass;
	Integer count;
	Double weight;
	Double performanceAggregation;
	boolean postprocessingIsStrongestEdgeEnteringAnActivity;
	boolean postprocessingIsStrongestEdgeExitingAnActivity;
	List<EventEdge> associatedEventEdges;
	Set<XOCObject> associatedObjects;
	
	public ModelEdge(XOCActivity sourceActivity, XOCActivity targetActivity, XOCObjectClass relationClass, List<EventEdge> associatedEventEdges) {
		this.sourceActivity = sourceActivity;
		this.targetActivity = targetActivity;
		this.relationClass = relationClass;
		this.count = 0;
		this.weight = 0.0;
		this.performanceAggregation = 0.0;
		this.postprocessingIsStrongestEdgeEnteringAnActivity = false;
		this.postprocessingIsStrongestEdgeExitingAnActivity = false;
		this.associatedEventEdges = associatedEventEdges;
		this.associatedObjects = new HashSet<XOCObject>();
		for (EventEdge edge : this.associatedEventEdges) {
			this.associatedObjects.add(edge.getRelationObject());
		}
	}
	
	public List<EventEdge> getAssociatedEventEdges() {
		return this.associatedEventEdges;
	}
	
	public Set<XOCObject> getAssociatedObjects() {
		return this.associatedObjects;
	}
	
	public XOCActivity getSourceActivity() {
		return this.sourceActivity;
	}
	
	public XOCActivity getTargetActivity() {
		return this.targetActivity;
	}
	
	public XOCObjectClass getRelationClass() {
		return this.relationClass;
	}
	
	public Integer getCount() {
		return this.count;
	}
	
	public void setCount(Integer count) {
		this.count = count;
	}
	
	public Double getWeight() {
		return this.weight;
	}
	
	public void setWeight(Double weight) {
		this.weight = weight;
	}
	
	public Double getPerformanceAggregation() {
		return this.performanceAggregation;
	}
	
	public void setPerformanceAggregation(Double performance) {
		this.performanceAggregation = performance;
	}
	
	public void reviseWeight() {
		double intermediate = this.weight / ModelConstants.weightThreshold;
		boolean isStrongestEdgeEnteringAnActivity = this.isStrongestEdgeEnteringAnActivity();
		boolean isStrongestEdgeExitingAnActivity = this.isStrongestEdgeExitingAnActivity();
		
		if (isStrongestEdgeEnteringAnActivity || isStrongestEdgeExitingAnActivity) {
			intermediate += ModelConstants.weightThreshold;
		}
		//Double oldWeight = new Double(this.weight);
		this.weight = intermediate / (1.0 + intermediate);
		//System.out.println("weight revised from "+oldWeight+" to "+this.weight);
	}
	
	public boolean isStrongestEdgeEnteringAnActivity() {
		return this.postprocessingIsStrongestEdgeEnteringAnActivity;
	}
	
	public void setStrongestEdgeEnteringAnActivity() {
		this.postprocessingIsStrongestEdgeEnteringAnActivity = true;
	}
	
	public boolean isStrongestEdgeExitingAnActivity() {
		return this.postprocessingIsStrongestEdgeExitingAnActivity;
	}
	
	public void setStrongestEdgeExitingAnActivity() {
		this.postprocessingIsStrongestEdgeExitingAnActivity = true;
	}
	
	public String getResume() {
		StringBuilder ret = new StringBuilder();
		ret.append("model edge from "+this.sourceActivity.getActivityLabel()+" to "+this.targetActivity.getActivityLabel()+" (perspective: "+this.getRelationClass().getClassLabel()+"; count: "+this.getCount()+"; weight: "+this.getWeight()+"; performance: "+this.getPerformanceAggregation()+")\n");
		return ret.toString();
	}
	
	public void printResume() {
		System.out.println(getResume());
	}
	
	public String convertTimeDiffToHumanReadable(double timeDiff0) {
		int timeDiff = (int)timeDiff0;
		int HALFMONTH = 1296000;
		int HALFDAY = 43200;
		int HALFHOUR = 1800;
		int HALFMINUTE = 30;
		
		if (timeDiff >= 2*HALFMONTH) {
			int noOfMonths = timeDiff / (2*HALFMONTH);
			int remaining = timeDiff - noOfMonths * 2*HALFMONTH;
			if (remaining > HALFMONTH) {
				noOfMonths+=1;
			}
			return noOfMonths + "M";
		}
		else if (timeDiff >= 2*HALFDAY) {
			int noOfDays = timeDiff / (2*HALFDAY);
			int remaining = timeDiff - noOfDays * 2*HALFDAY;
			if (remaining > HALFDAY) {
				noOfDays += 1;
			}
			return noOfDays + "D";
		}
		else if (timeDiff >= 2*HALFHOUR) {
			int noOfHours = timeDiff / (2*HALFHOUR);
			int remaining = timeDiff - noOfHours * 2*HALFHOUR;
			if (remaining > HALFHOUR) {
				noOfHours += 1;
			}
			return noOfHours + "h";
		}
		else if (timeDiff >= 2*HALFMINUTE) {
			int noOfMinutes = timeDiff / (2*HALFMINUTE);
			int remaining = timeDiff - noOfMinutes * 2*HALFMINUTE;
			if (remaining > HALFMINUTE) {
				noOfMinutes += 1;
			}
			return noOfMinutes + "m";
		}
		
		return String.format("%.0f", timeDiff0)+"s";
	}
	
	public double getPerspectiveValue(String perspective) {
		if (perspective.equals("count")) {
			return this.getCount();
		}
		else if (perspective.equals("weight")) {
			return this.getWeight();
		}
		else if (perspective.equals("type")) {
			return this.getCount();
		}
		else if (perspective.equals("performance")) {
			return this.getPerformanceAggregation();
		}
		return 0.0;
	}
	
	public String getEdgeLabel(String perspective) {
		if (perspective.equals("count")) {
			return this.getRelationClass().getClassLabel() + " (" + this.getCount() + ")";
		}
		else if (perspective.equals("weight")) {
			return this.getRelationClass().getClassLabel() + " (" + String.format("%.2f", this.getWeight()) + ")";
		}
		else if (perspective.equals("type")) {
			return this.getRelationClass().getClassLabel();
		}
		else if (perspective.equals("performance")) {
			return this.relationClass.getClassLabel() + " (" + convertTimeDiffToHumanReadable(this.getPerformanceAggregation()) + ")";
			//return this.relationClass.getClassLabel() + " (" + String.format("%.0f", this.getPerformanceAggregation())+"s)";
		}
		return "";
	}
	
	@Override
	public String toString() {
		return this.sourceActivity.getActivityLabel()+"->"+this.targetActivity.getActivityLabel()+" ("+this.relationClass.getClassLabel()+")";
	}
}
