package org.processmining.starstar.model;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.processmining.starstar.elements.XOCActivity;
import org.processmining.starstar.elements.XOCObject;
import org.processmining.starstar.parser.ParsedXOC;

public class Subgraph {
	ParsedXOC parsedXOC;
	public Set<String> involvedObjectNames;
	public Set<ModelEdge> involvedEdges;
	public Set<XOCActivity> involvedActivitiesInEdges;
	public Set<XOCActivity> involvedSourceActivities;
	public Set<XOCActivity> involvedTargetActivities;
	public Set<String> involvedObjectsClassesInEdges;
	public String graphOperator;
	public String graphBasicDirection;
	public Integer involvedActivitiesMinCount;
	public Integer involvedActivitiesMaxCount;
	public Double involvedActivitiesAvgCount;
	public Double score;
	
	public Subgraph(ParsedXOC parsedXOC) {
		this.parsedXOC = parsedXOC;
		involvedEdges = new HashSet<ModelEdge>();
		involvedActivitiesInEdges = new HashSet<XOCActivity>();
		involvedSourceActivities = new HashSet<XOCActivity>();
		involvedTargetActivities = new HashSet<XOCActivity>();
		involvedObjectsClassesInEdges = new HashSet<String>();
		graphOperator = "AND";
		graphBasicDirection = "forward";
		involvedObjectNames = new HashSet<String>();
		involvedActivitiesMinCount = 99999999;
		involvedActivitiesMaxCount = 0;
		involvedActivitiesAvgCount = new Double(0.0);
		score = 0.0;
	}
	
	public Subgraph(Subgraph otherGraph) {
		parsedXOC = otherGraph.parsedXOC;
		involvedEdges = new HashSet<ModelEdge>(otherGraph.involvedEdges);
		involvedActivitiesInEdges = new HashSet<XOCActivity>(otherGraph.involvedActivitiesInEdges);
		involvedSourceActivities = new HashSet<XOCActivity>(otherGraph.involvedSourceActivities);
		involvedTargetActivities = new HashSet<XOCActivity>(otherGraph.involvedTargetActivities);
		involvedObjectsClassesInEdges = new HashSet<String>(otherGraph.involvedObjectsClassesInEdges);
		graphOperator = otherGraph.graphOperator;
		graphBasicDirection = otherGraph.graphBasicDirection;
		involvedObjectNames = new HashSet<String>();
		involvedActivitiesMinCount = 99999999;
		involvedActivitiesMaxCount = 0;
		involvedActivitiesAvgCount = new Double(0.0);
		score = 0.0;
	}
	
	public void setGraphOperator(String edgesOperator) {
		this.graphOperator = edgesOperator;
	}
	
	public void setGraphBasicDirection(String graphBasicDirection) {
		this.graphBasicDirection = graphBasicDirection;
	}
	
	public void addEdge(ModelEdge edge) {
		this.involvedEdges.add(edge);
		this.involvedActivitiesInEdges.add(edge.getSourceActivity());
		this.involvedSourceActivities.add(edge.getSourceActivity());
		this.involvedActivitiesInEdges.add(edge.getTargetActivity());
		this.involvedTargetActivities.add(edge.getTargetActivity());
		this.involvedObjectsClassesInEdges.add(edge.getRelationClass().getClassLabel());
		
		double sumOfActivitiesCount = 0.0;
		for (XOCActivity activity : this.involvedActivitiesInEdges) {
			Integer activityCount = activity.getActivityCount();
			sumOfActivitiesCount += activityCount;
			if (activityCount < involvedActivitiesMinCount) {
				involvedActivitiesMinCount = activityCount;
			}
			if (activityCount > involvedActivitiesMaxCount) {
				involvedActivitiesMaxCount = activityCount;
			}
		}
		if (this.involvedActivitiesInEdges.size() > 0) {
			this.involvedActivitiesAvgCount = sumOfActivitiesCount / this.involvedActivitiesInEdges.size();
		}
	}
	
	public void calculateInvolvedEdgesAndScore() {
		involvedObjectNames = new HashSet<String>();
		for (ModelEdge edge : involvedEdges) {
			Set<String> thisPiece = parsedXOC.getObjectsInvolvedInEdge(edge);
			if (involvedObjectNames.size() == 0) {
				involvedObjectNames = thisPiece;
			}
			else {
				if (graphOperator == "AND") {
					involvedObjectNames.retainAll(thisPiece);
				}
				else if (graphOperator == "XOR") {
					involvedObjectNames.addAll(thisPiece);
				}
			}
		}
		double operatorMultiplier = 1.0;
		if (graphOperator == "AND") {
			operatorMultiplier = 2.0;
		}
		
		this.score = (double) (this.involvedEdges.size() * this.involvedObjectsClassesInEdges.size() * this.involvedObjectNames.size() * operatorMultiplier);
	}
	
	public void removeByLexicographicOrder(List<ModelEdge> possibleExpansionEdges) {
		// remove possible expansion edges based on source and target activity comparison
		// and lexicographic order
		int i = 0;
		while (i < possibleExpansionEdges.size()) {
			boolean mustContinue = false;
			ModelEdge edge = possibleExpansionEdges.get(i);
			for (ModelEdge otherEdge : this.involvedEdges) {
				if (edge.getSourceActivity().equals(otherEdge.getSourceActivity())) {
					String thisEdgeTargetActivityLabel = edge.getTargetActivity().getActivityLabel();
					String otherEdgeTargetActivityLabel = otherEdge.getTargetActivity().getActivityLabel();
					if (thisEdgeTargetActivityLabel.compareTo(otherEdgeTargetActivityLabel) < 1) {
						//System.out.println("thisEdgeTargetActivityLabel = "+thisEdgeTargetActivityLabel+" otherEdgeTargetActivityLabel="+otherEdgeTargetActivityLabel);
						possibleExpansionEdges.remove(i);
						mustContinue = true;
						break;
					}
				}
				if (edge.getTargetActivity().equals(otherEdge.getTargetActivity())) {
					String thisEdgeSourceActivityLabel = edge.getSourceActivity().getActivityLabel();
					String otherEdgeSourceActivityLabel = otherEdge.getSourceActivity().getActivityLabel();
					if (thisEdgeSourceActivityLabel.compareTo(otherEdgeSourceActivityLabel) < 1) {
						//System.out.println("thisEdgeSourceActivityLabel = "+thisEdgeSourceActivityLabel+" otherEdgeSourceActivityLabel="+otherEdgeSourceActivityLabel);
						possibleExpansionEdges.remove(i);
						mustContinue = true;
						break;
					}
				}
			}
			if (mustContinue) {
				continue;
			}
			i++;
		}
	}
	
	public void removeEdgesThatHaveLittleInCommon(List<ModelEdge> possibleExpansionEdges) {
		// remove possible expansion edges that have little in common with the objects currently involved in the subgraph
		int i = 0;
		while (i < possibleExpansionEdges.size()) {
			ModelEdge edge = possibleExpansionEdges.get(i);
			Set<String> edgeInvolvedObjectsNames = new HashSet<String>();
			for (XOCObject object : edge.getAssociatedObjects()) {
				edgeInvolvedObjectsNames.add(object.getObjectName());
			}
			edgeInvolvedObjectsNames.retainAll(this.involvedObjectNames);
			if (edgeInvolvedObjectsNames.size() > 0.5 * this.involvedObjectNames.size()) {
				
			}
			else {
				possibleExpansionEdges.remove(i);
				continue;
			}
			i++;
		}
	}
	
	public void removeEdgesThatHaveTooMuchInCommon(List<ModelEdge> possibleExpansionEdges) {
		int i = 0;
		while (i < possibleExpansionEdges.size()) {
			ModelEdge edge = possibleExpansionEdges.get(i);
			Set<String> edgeInvolvedObjectNames = new HashSet<String>();
			for (XOCObject object : edge.getAssociatedObjects()) {
				edgeInvolvedObjectNames.add(object.getObjectName());
			}
			edgeInvolvedObjectNames.retainAll(this.involvedObjectNames);
			if (edgeInvolvedObjectNames.size() < 0.05 * this.involvedObjectNames.size()) {
				
			}
			else {
				possibleExpansionEdges.remove(i);
				continue;
			}
			i++;
		}
	}
	
	public Set<ModelEdge> getPossibleExpansionEdgesAND(String direction) {
		Set<ModelEdge> possibleExpansionEdgesSet = new HashSet<ModelEdge>();
		for (ModelEdge edge : this.involvedEdges) {
			XOCActivity activity = null;
			List<ModelEdge> thisActivityExpansionEdges = new ArrayList<ModelEdge>();
			if (direction == "forwardForward" || direction == "forwardBackward") {
				activity = edge.getTargetActivity();
			}
			else if (direction == "backwardBackward" || direction == "backwardForward") {
				activity = edge.getSourceActivity();
			}
			if (direction == "forwardForward" || direction == "backwardForward") {
				thisActivityExpansionEdges = activity.getOutputModelEdges();
			}
			else if (direction == "backwardBackward" || direction == "fowardBackward") {
				thisActivityExpansionEdges = activity.getInputModelEdges();
			}
			possibleExpansionEdgesSet.addAll(thisActivityExpansionEdges);
		}
		possibleExpansionEdgesSet.removeAll(involvedEdges);
		
		List<ModelEdge> possibleExpansionEdges = new ArrayList<ModelEdge>(possibleExpansionEdgesSet);
		
		removeByLexicographicOrder(possibleExpansionEdges);
		removeEdgesThatHaveLittleInCommon(possibleExpansionEdges);
		
		Set<ModelEdge> possibleExpansionEdgesFinalSet = new HashSet<ModelEdge>(possibleExpansionEdges);
		
		return possibleExpansionEdgesFinalSet;
	}
	
	public Set<ModelEdge> getPossibleExpansionEdgesOR() {
		Set<ModelEdge> possibleExpansionEdgesSet = new HashSet<ModelEdge>();
		
		if (this.involvedEdges.size() == 1 || (this.graphOperator == "XOR" && (this.involvedSourceActivities.size() == 1 || this.involvedTargetActivities.size() == 1))) {
			if (this.involvedSourceActivities.size() == 1) {
				List<XOCActivity> sActivities = new ArrayList<XOCActivity>(this.involvedSourceActivities);
				XOCActivity activity = sActivities.get(0);
				for (ModelEdge edge : activity.getOutputModelEdges()) {
					if (!this.involvedEdges.contains(edge) && this.involvedObjectsClassesInEdges.contains(edge.getRelationClass().getClassLabel())) {
						possibleExpansionEdgesSet.add(edge);
					}
				}
			}
			if (this.involvedTargetActivities.size() == 1) {
				List<XOCActivity> eActivities = new ArrayList<XOCActivity>(this.involvedTargetActivities);
				XOCActivity activity = eActivities.get(0);
				for (ModelEdge edge : activity.getInputModelEdges()) {
					if (!this.involvedEdges.contains(edge) && this.involvedObjectsClassesInEdges.contains(edge.getRelationClass().getClassLabel())) {
						possibleExpansionEdgesSet.add(edge);
					}
				}
			}
		}
		
		List<ModelEdge> possibleExpansionEdges = new ArrayList<ModelEdge>(possibleExpansionEdgesSet);
		
		removeByLexicographicOrder(possibleExpansionEdges);
		removeEdgesThatHaveTooMuchInCommon(possibleExpansionEdges);
		
		Set<ModelEdge> possibleExpansionEdgesFinalSet = new HashSet<ModelEdge>(possibleExpansionEdges);
		
		return possibleExpansionEdgesFinalSet;
	}
	
	public void printResume() {
		System.out.println("\n\nSubgraph involvedActivitiesInEdges = "+involvedActivitiesInEdges);
		System.out.println("involvedObjectsClassesInEdges = "+involvedObjectsClassesInEdges);
		System.out.println("involvedEdges = "+involvedEdges);
		System.out.println("graphOperator = "+graphOperator);
		System.out.println("graphBasicDirection = "+graphBasicDirection);
		System.out.println("involvedObjectNames size = "+involvedObjectNames.size());
		System.out.println("involvedActivitiesMinCount = "+involvedActivitiesMinCount);
		System.out.println("involvedActivitiesMaxCount = "+involvedActivitiesMaxCount);
		System.out.println("involvedActivitiesAvgCount = "+involvedActivitiesAvgCount);
		System.out.println("score = "+score);
	}
	
	public String getShortResume() {
		StringBuilder ret = new StringBuilder();
		
		ret.append("Subgraph ("+graphOperator+") with score = "+score);
		
		for (ModelEdge edge : this.involvedEdges) {
			ret.append("\n"+edge.toString());
		}
		
		return ret.toString();
	}
	
	@Override
	public String toString() {
		return "Subgraph ("+graphOperator+") with score = "+score;
	}
}
