package org.processmining.starstar.plugins;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.neo4j.driver.v1.AuthTokens;
import org.neo4j.driver.v1.Driver;
import org.neo4j.driver.v1.GraphDatabase;
import org.neo4j.driver.v1.Record;
import org.neo4j.driver.v1.Session;
import org.neo4j.driver.v1.Transaction;
import org.neo4j.driver.v1.TransactionWork;
import org.processmining.starstar.elements.XOCActivity;
import org.processmining.starstar.elements.XOCEvent;
import org.processmining.starstar.elements.XOCObject;
import org.processmining.starstar.elements.XOCObjectClass;
import org.processmining.starstar.model.EventEdge;
import org.processmining.starstar.model.ModelEdge;

public class StarStarModelNeo4JImporter  implements AutoCloseable {
    private final Driver driver;
    Map<String, XOCActivity> activities;
    Map<String, XOCObjectClass> objectClasses;
    Map<String, XOCObject> allObjects;
    Map<String, XOCEvent> allEvents;
    Map<XOCObject, List<XOCObject>> intersectedObjects;
    
    public StarStarModelNeo4JImporter(String uri, String user, String password) {
    	this.activities = new HashMap<String, XOCActivity>();
    	this.objectClasses = new HashMap<String, XOCObjectClass>();
    	this.allObjects = new HashMap<String, XOCObject>();
    	this.allEvents = new HashMap<String, XOCEvent>();
    	this.intersectedObjects = new HashMap<XOCObject, List<XOCObject>>();
    	this.driver = GraphDatabase.driver( uri, AuthTokens.basic( user, password ) );
    }

    @Override
    public void close() throws Exception
    {
        driver.close();
    }
    
    public List<XOCActivity> getActivities() {
    	return new ArrayList<XOCActivity>(this.activities.values());
    }
    
    public Map<XOCObject, List<XOCObject>> returnIntersectedObjects() {
    	return this.intersectedObjects;
    }
    
    public Map<String, XOCObjectClass> getXocObjectClasses() {
    	return objectClasses;
    }
    
    public void getActivitiesFromNeo4J() {
    	if (activities.keySet().size() == 0) {
	        try ( Session session = driver.session() )
	        {
	        	List<Record> activitiesInNeo4J_Left
	        	= session.readTransaction(new TransactionWork<List<Record>>() {
					public List<Record> execute(Transaction tx) {
						List<Record> results = tx.run("MATCH (a)-[r:ENTITYFOLLOW]->(b) RETURN a").list();	
						return results;
					}
	        	});
	        	
	        	List<Record> activitiesInNeo4J_Right = session.readTransaction(new TransactionWork<List<Record>>() {
					public List<Record> execute(Transaction tx) {
						// TODO Auto-generated method stub
						List<Record> results = tx.run("MATCH (a)-[r:ENTITYFOLLOW]->(b) RETURN b").list();	
						return results;
					}
	        	});
	        	
	        	for (Record r : activitiesInNeo4J_Left) {
	        		String activityName = r.get("a").get("customEntityName").toString().replaceAll("\"", "").split(" \\(")[0].toLowerCase();
	        		XOCActivity activity = new XOCActivity(activityName);
	        		activities.put(activityName, activity);
	        	}
	        	
	        	for (Record r : activitiesInNeo4J_Right) {
	        		String activityName = r.get("b").get("customEntityName").toString().replaceAll("\"", "").split(" \\(")[0].toLowerCase();
	        		XOCActivity activity = new XOCActivity(activityName);
	        		activities.put(activityName, activity);
	        	}
	        	
	        	List<Record> modelEdgesInNeo4J = session.readTransaction(new TransactionWork<List<Record>> () {
					public List<Record> execute(Transaction tx) {
						List<Record> result = tx.run("MATCH (a)-[r:ENTITYFOLLOW]->(b) RETURN toLower(a.customEntityName), toLower(b.customEntityName), r.relCauseType, r.relationWeight, r.count, r.averageTime").list();
						return result;
					}
	        	});
	        	
	        	List<Record> activitiesCountInNeo4J = session.readTransaction(new TransactionWork<List<Record>> () {
	        		public List<Record> execute(Transaction tx) {
	        			List<Record> result = tx.run("MATCH (a) WHERE (a.customEntityType <> \"master\" OR a.customEntityLevel <> \"master\") RETURN toLower(a.customEntityName), count(*)").list();
	        			return result;
	        		}
	        	});
	        	
	        	for (Record r : activitiesCountInNeo4J) {
	        		String actName = r.get("toLower(a.customEntityName)").toString().replaceAll("\"", "");
	        		Integer activityCount = Integer.parseInt(r.get("count(*)").toString());
	
	        		if (activities.containsKey(actName)) {
	        			activities.get(actName).setActivityCount(activityCount);
	        		}
	        	}
	        	
	        	List<String> alreadyAdded = new ArrayList<String>();
	        	for (Record r : modelEdgesInNeo4J) {
	        		String firstEntityName = r.get("toLower(a.customEntityName)").toString().replaceAll("\"", "").split(" \\(")[0];
	        		String secondEntityName = r.get("toLower(b.customEntityName)").toString().replaceAll("\"", "").split(" \\(")[0];
	        		String relCauseType = r.get("r.relCauseType").toString().replaceAll("\"", "");
	        		
	        		String concatenation = firstEntityName + "@@@" + secondEntityName + "@@@" + relCauseType;
	        		
	        		if (!alreadyAdded.contains(concatenation)) {
	            		Integer count = Integer.parseInt(r.get("r.count").toString());
	            		Double relationWeight = Double.parseDouble(r.get("r.relationWeight").toString());
	            		Double averageTime = 0.0;
	            		try {
	            			averageTime = Double.parseDouble(r.get("r.averageTime").toString());
	            		}
	            		catch (Exception ex) {
	            			//ex.printStackTrace();
	            		}
	            		if (!objectClasses.containsKey(relCauseType)) {
	            			objectClasses.put(relCauseType, new XOCObjectClass(relCauseType));
	            		}
	            		
	            		if (activities.containsKey(firstEntityName) && activities.containsKey(secondEntityName)) {
	            			ModelEdge edge = new ModelEdge(activities.get(firstEntityName), activities.get(secondEntityName), objectClasses.get(relCauseType), new ArrayList<EventEdge>());
	            			edge.setCount(count);
	            			edge.setWeight(relationWeight);
	            			edge.setPerformanceAggregation(averageTime);
	            			edge.printResume();
	            			activities.get(firstEntityName).addOutputModelEdge(edge);
	            			activities.get(secondEntityName).addInputModelEdge(edge);
	            		}
	            		
	            		alreadyAdded.add(concatenation);
	            		
	            		/*System.out.println("firstEntityName="+firstEntityName);
	            		System.out.println("secondEntityName="+secondEntityName);
	            		System.out.println("relCauseType="+relCauseType);
	            		System.out.println("relationWeight="+relationWeight);
	            		System.out.println("averageTime="+averageTime);*/
	        		}
	        	}
	        }
    	}
    }
    
    public void getAllObjectsFromNeo4J() {
	    if (allObjects.keySet().size() == 0) {
	        try ( Session session = driver.session() )
	        {
	        	List<Record> allObjectRecords = session.readTransaction(new TransactionWork<List<Record>> () {
	
					public List<Record> execute(Transaction tx) {
						// TODO Auto-generated method stub
						List<Record> allObjectRecords = tx.run("MATCH (a)-[r:EVENTLINKED]->(b) RETURN b.customEntityId, toLower(b.customEntityName)").list();
						return allObjectRecords;
					}
	        	});
	
	        	
	        	for (Record r : allObjectRecords) {
	        		String objectId = r.get("b.customEntityId").toString().replaceAll("\"", "");
	        		String objectName = r.get("toLower(b.customEntityName)").toString().replaceAll("\"", "").split(" \\(")[0];
	        		
	        		//System.out.println(objectId+" "+objectName);
	        		
	        		if (!allObjects.containsKey(objectId) && objectClasses.containsKey(objectName)) {
	        			XOCObject newObject = new XOCObject(objectId, objectClasses);
	        			XOCObjectClass objectClass = objectClasses.get(objectName);
	        			newObject.setObjectClass(objectClass);
	        			objectClass.subscribeObject(newObject);
	        			        			
	        			allObjects.put(objectId, newObject);
	        		}
	        	}
	        }
    	}
    }
    
    public void getAllEventsFromNeo4J() {
    	if (allEvents.keySet().size() == 0) {
	        try ( Session session = driver.session() )
	        {
	        	List<Record> allEventRecords = session.readTransaction(new TransactionWork<List<Record>> () {
	
					public List<Record> execute(Transaction tx) {
						// TODO Auto-generated method stub
						List<Record> allEventRecords = tx.run("MATCH (b)-[r:EVENTLINKED]->(a) RETURN b.customEntityId, toLower(b.customEntityName), a.customEntityId, toLower(a.customEntityName), b.customEntityTimestamp, a.customEntityTimestamp").list();
						return allEventRecords;
					}
	        	});
	        	
	        	for (Record r : allEventRecords) {
	        		String firstEntityId = r.get("b.customEntityId").toString().replaceAll("\"", "");
	        		String firstEntityName = r.get("toLower(b.customEntityName)").toString().replaceAll("\"", "").split(" \\(")[0];
	        		String secondEntityId = r.get("a.customEntityId").toString().replaceAll("\"", "");
	        		String secondEntityName = r.get("toLower(a.customEntityName)").toString().replaceAll("\"", "").split(" \\(")[0];
	        		String firstEntityTimestamp = r.get("b.customEntityTimestamp").toString();
	        		String secEntityTimestamp = r.get("a.customEntityTimestamp").toString();
	        		
	        		if (firstEntityTimestamp != "NULL" && activities.containsKey(firstEntityName) && allObjects.containsKey(secondEntityId)) {
	        			Double unixTimest = new Double(firstEntityTimestamp);
	        			long unixTimestTransf = (long)(double)unixTimest * 1000;
	        			Date date = new Date(unixTimestTransf);
	        			XOCActivity activity = activities.get(firstEntityName);
	        			
	        			XOCEvent newEvent = null;
	        			if (!allEvents.containsKey(firstEntityId)) {
	        				newEvent = new XOCEvent(firstEntityId, date, activity);
		        			allEvents.put(firstEntityId, newEvent);
		        		}
	        			else {
	        				newEvent = allEvents.get(firstEntityId);
	        			}
	
	        			allObjects.get(secondEntityId).addSubscribedEvent(newEvent);
	        			activities.get(firstEntityName).addSubscribedEvent(newEvent);
	        		}
	        		
	        		if (secEntityTimestamp != "NULL" && activities.containsKey(secondEntityName) && allObjects.containsKey(secondEntityId)) {
	        			Double unixTimest = new Double(secEntityTimestamp);
	        			long unixTimestTransf = (long)(double)unixTimest * 1000;
	        			Date date = new Date(unixTimestTransf);
	        			XOCActivity activity = activities.get(secondEntityName);
	        			
	        			XOCEvent newEvent = null;
	        			if (!allEvents.containsKey(secondEntityId)) {
	        				newEvent = new XOCEvent(secondEntityId, date, activity);
	        				allEvents.put(secondEntityId, newEvent);
	        			}
	        			else {
	        				newEvent = allEvents.get(secondEntityId);
	        			}
	        			
	        			allObjects.get(secondEntityId).addSubscribedEvent(newEvent);
	        			activities.get(secondEntityName).addSubscribedEvent(newEvent);
	        		}
	        		
	        		//XOCEvent newEvent = new XOCEvent(secondEntityId);
	        	}
	        }
    	}
    }
    
	public void calculateObjectIntersection() {
		for (XOCObject o1 : this.allObjects.values()) {
			this.intersectedObjects.put(o1, new ArrayList<XOCObject>());
			List<XOCEvent> e1 = new ArrayList<XOCEvent>(o1.getSubscribedEvents());
			if (e1.size() > 0) {
				for (XOCObject o2 : this.allObjects.values()) {
					if (!o1.getObjectName().equals(o2.getObjectName())) {
						List<XOCEvent> e2 = new ArrayList<XOCEvent>(o2.getSubscribedEvents());
						if (e2.size() > 0) {
							e1 = new ArrayList<XOCEvent>(o1.getSubscribedEvents());

							e1.retainAll(e2);
							if (e1.size() > 0) {
								this.intersectedObjects.get(o1).add(o2);
							}
						}
					}
				}
			}
			
			if (this.intersectedObjects.get(o1).size() > 0) {

			}
		}
	}

	public static void main(String[] args) {
		StarStarModelNeo4JImporter importer = new StarStarModelNeo4JImporter("bolt://localhost:7687", "neo4j", "aaaaa");
		importer.getActivitiesFromNeo4J();
		importer.getAllObjectsFromNeo4J();
		System.out.println(importer.allObjects.keySet().size());
		importer.getAllEventsFromNeo4J();
		System.out.println(importer.allEvents.keySet().size());
		importer.calculateObjectIntersection();
		
		System.out.println("finished");
	}
}
